import io
from typing import Dict, List

import numpy as np

from jina.executors.segmenters import BaseSegmenter


class PDFExtractorSegmenter(BaseSegmenter):
    """
    :class:`PDFExtractorSegmenter` Extracts data (text and images) from PDF files.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def segment(self, uri: str, buffer: bytes, *args, **kwargs) -> List[Dict]:
        """
        Segements PDF files. Extracts data from them.

        Checks if the input is a string of the filename,
        or if it's the file in bytes.
        It will then extract the data from the file, creating a list for images,
        and text.

        :param uri: File name of PDF
        :type uri: str
        :param buffer: PDF file in bytes
        :type buffer: bytes
        :returns: A list of documents with the extracted data
        :rtype: List[Dict]
        """

        import fitz
        import PyPDF2

        if uri:
            pdf_img = fitz.open(uri)
            pdf_text = open(uri, 'rb')
        elif buffer:
            pdf_img = fitz.open(stream=buffer, filetype='pdf')
            pdf_text = io.BytesIO(buffer)
        else:
            raise ValueError('No value found in "buffer" or "uri"')

        chunks = []
        # Extract images
        with pdf_img:
            for page in range(len(pdf_img)):
                for img in pdf_img.getPageImageList(page):
                    xref = img[0]
                    pix = fitz.Pixmap(pdf_img, xref)
                    np_arr = np.frombuffer(pix.samples, dtype=np.uint8).reshape(pix.h, pix.w, pix.n).astype('float32')
                    if pix.n - pix.alpha < 4:  # if gray or RGB
                        chunks.append(
                            dict(blob=np_arr, weight=1.0, mime_type='image/png'))
                    else:  # if CMYK:
                        pix = fitz.Pixmap(fitz.csRGB, pix)  # Convert to RGB
                        np_arr = np.frombuffer(pix.samples, dtype=np.uint8).reshape(pix.h, pix.w, pix.n).astype(
                            'float32')
                        chunks.append(
                            dict(blob=np_arr, weight=1.0, mime_type='image/png'))

        # Extract text
        with pdf_text:
            text = ""
            pdf_reader = PyPDF2.PdfFileReader(pdf_text)
            count = pdf_reader.numPages
            for page in range(count):
                page = pdf_reader.getPage(page)
                text += page.extractText()
            if text:
                chunks.append(
                    dict(text=text, weight=1.0, mime_type='text/plain'))

        return chunks



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from jina.executors.encoders.numeric import TransformEncoder

class FeatureAgglomerationEncoder(TransformEncoder):
    """
    Encode data agglomerating features.

    Recursively merges features that minimally increases a given linkage distance.
    Similar to AgglomerativeClustering, but recursively merges features instead of samples.

    Encodes data from an ndarray in size `B x T` into an ndarray in size `B x D`
    Where `B` is the batch's size and `T` and `D` are the dimensions pre (`T`)
    and after (`D`) the encoding.

    See more `at <https://scikit-learn.org/stable/modules/generated/sklearn.cluster.FeatureAgglomeration.html>`_
    """

    def post_init(self):
        """Load FeatureAgglomeration model"""
        super().post_init()
        if not self.model:
            from sklearn.cluster import FeatureAgglomeration
            self.model = FeatureAgglomeration(n_clusters=self.output_dim)



import inspect
from typing import Dict, List

from .. import BaseExecutor
from ...helper import typename


class BaseSegmenter(BaseExecutor):
    """:class:`BaseSegmenter` works on doc-level,
    it chunks Documents into set of Chunks"""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super().__init__(*args, **kwargs)
        self.required_keys = {
            k for k in inspect.getfullargspec(self.segment).args if k != 'self'
        }
        if not self.required_keys:
            self.logger.warning(
                f'{typename(self)} works on keys, but no keys are specified'
            )

    def segment(self, *args, **kwargs) -> List[Dict]:
        """
        :return: a list of chunks-level info represented by a dict
        """
        raise NotImplementedError



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional

from . import FlatRecursiveMixin, BaseExecutableDriver

if False:
    from .. import DocumentSet


class CraftDriver(FlatRecursiveMixin, BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`craft` by default """

    def __init__(
        self, executor: Optional[str] = None, method: str = 'craft', *args, **kwargs
    ):
        super().__init__(executor, method, *args, **kwargs)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs):
        for doc in docs:
            _args_dict = doc.get_attrs(*self.exec.required_keys)
            ret = self.exec_fn(**_args_dict)
            if ret:
                doc.set_attrs(**ret)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterable, Tuple

from .. import QuerySetReader, ContextAwareRecursiveMixin, BaseRecursiveDriver

if False:
    from ...types.sets import DocumentSet


class ReverseQL(QuerySetReader, ContextAwareRecursiveMixin, BaseRecursiveDriver):
    """Reverses the order of the provided ``docs``.

    This is often useful when the proceeding Pods require only a signal, not the full message.

    Example ::
    - !Chunk2DocRankerDriver {}
    - !ReverseQL {}

    will reverse the order of the documents returned by the `Chunk2DocRankerDriver` before sending them to the next `Pod`
    """

    def __init__(self, traversal_paths: Tuple[str] = ('r',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)

    def _apply_all(
        self, doc_sequences: Iterable['DocumentSet'], *args, **kwargs
    ) -> None:
        for docs in doc_sequences:
            docs.reverse()



from typing import Type

from . import BaseNdArray
from .dense import BaseDenseNdArray
from .dense.numpy import DenseNdArray
from .sparse import BaseSparseNdArray
from .sparse.scipy import SparseNdArray
from ...proto import jina_pb2

__all__ = ['NdArray']


class NdArray(BaseNdArray):
    """
    :class:`NdArray` is one of the **primitive data type** in Jina.

    It offers a Pythonic interface to allow users access and manipulate
    :class:`jina.jina_pb2.NdArrayProto` object without working with Protobuf itself.

    A generic view of the Protobuf NdArray, unifying the view of DenseNdArray and SparseNdArray

    This class should be used in nearly all the Jina context.

    Simple usage:

    .. highlight:: python
    .. code-block:: python

        # start from empty proto
        a = NdArray()

        # start from an existig proto
        a = NdArray(doc.embedding)

        # set value
        a.value = np.random.random([10, 5])

        # get value
        print(a.value)

        # set value to a TF sparse tensor
        a.is_sparse = True
        a.value = SparseTensor(...)
        print(a.value)

    Advanced usage:

    :class:`NdArray` also takes a dense NdArray and a sparse NdArray constructor
    as arguments. You can consider them as the backend for dense and sparse NdArray. The combination
    is your choice, it could be:

    .. highlight:: python
    .. code-block:: python

        # numpy (dense) + scipy (sparse)
        from .dense.numpy import DenseNdArray
        from .sparse.scipy import SparseNdArray
        NdArray(dense_cls=DenseNdArray, sparse_cls=SparseNdArray)

        # numpy (dense) + pytorch (sparse)
        from .dense.numpy import DenseNdArray
        from .sparse.pytorch import SparseNdArray
        NdArray(dense_cls=DenseNdArray, sparse_cls=SparseNdArray)

        # numpy (dense) + tensorflow (sparse)
        from .dense.numpy import DenseNdArray
        from .sparse.tensorflow import SparseNdArray
        NdArray(dense_cls=DenseNdArray, sparse_cls=SparseNdArray)

    Once you set `sparse_cls`, it will only accept the data type in that particular type.
    That is, you can not use a :class:`NdArray` equipped with Tensorflow sparse to
    set/get Pytorch or Scipy sparse matrices.

    """

    def __init__(
        self,
        proto: 'jina_pb2.NdArrayProto' = None,
        is_sparse: bool = False,
        dense_cls: Type['BaseDenseNdArray'] = DenseNdArray,
        sparse_cls: Type['BaseSparseNdArray'] = SparseNdArray,
        *args,
        **kwargs
    ):
        """

        :param proto: the protobuf message, when not given then create a new one via :meth:`get_null_proto`
        :param is_sparse: if the ndarray is sparse, can be changed later
        :param dense_cls: the to-be-used class for DenseNdArray when `is_sparse=False`
        :param sparse_cls: the to-be-used class for SparseNdArray when `is_sparse=True`
        :param args:
        :param kwargs:
        """
        self.is_sparse = is_sparse
        self.dense_cls = dense_cls
        self.sparse_cls = sparse_cls
        super().__init__(proto, *args, **kwargs)
        self._args = args
        self._kwargs = kwargs

    def null_proto(self):
        """Get the new protobuf representation."""
        return jina_pb2.NdArrayProto()

    @property
    def value(self):
        """Get the value of protobuf and return in corresponding type."""
        stype = self._pb_body.WhichOneof('content')
        if stype == 'dense':
            return self.dense_cls(self._pb_body.dense).value
        elif stype == 'sparse':
            return self.sparse_cls(self._pb_body.sparse).value

    @value.setter
    def value(self, value):
        """Set the value of protobuf and with :param:`value`."""
        if self.is_sparse:
            self.sparse_cls(self._pb_body.sparse).value = value
        else:
            self.dense_cls(self._pb_body.dense).value = value

    def _build_content_dict(self):
        return {'value': self.value, 'is_sparse': self.is_sparse}



from collections.abc import MutableSequence
from typing import Union, Iterable, Tuple, Sequence, List

import numpy as np

from ...helper import typename
from ...logging import default_logger

try:
    # when protobuf using Cpp backend
    from google.protobuf.pyext._message import (
        RepeatedCompositeContainer as RepeatedContainer,
    )
except:
    # when protobuf using Python backend
    from google.protobuf.internal.containers import (
        RepeatedCompositeFieldContainer as RepeatedContainer,
    )

from ...proto.jina_pb2 import DocumentProto
from .traversable import TraversableSequence

if False:
    from ..document import Document

__all__ = ['DocumentSet']


class DocumentSet(TraversableSequence, MutableSequence):
    """
    :class:`DocumentSet` is a mutable sequence of :class:`Document`.
    It gives an efficient view of a list of Document. One can iterate over it like
    a generator but ALSO modify it, count it, get item, or union two 'DocumentSet's using the '+' and '+=' operators.

    :param docs_proto: A list of :class:`Document`
    :type docs_proto: Union['RepeatedContainer', Sequence['Document']]
    """

    def __init__(self, docs_proto: Union['RepeatedContainer', Sequence['Document']]):
        super().__init__()
        self._docs_proto = docs_proto
        self._docs_map = {}

    def insert(self, index: int, doc: 'Document') -> None:
        """
        Insert :param:`doc.proto` at :param:`index` into the list of `:class:`DocumentSet` .

        :param index: Position of the insertion.
        :param doc: The doc needs to be inserted.
        """
        self._docs_proto.insert(index, doc.proto)

    def __setitem__(self, key, value: 'Document'):
        if isinstance(key, int):
            self._docs_proto[key].CopyFrom(value)
        elif isinstance(key, str):
            self._docs_map[key].CopyFrom(value)
        else:
            raise IndexError(f'do not support this index {key}')

    def __delitem__(self, index):
        del self._docs_proto[index]

    def __len__(self):
        return len(self._docs_proto)

    def __iter__(self):
        from ..document import Document

        for d in self._docs_proto:
            yield Document(d)

    def __getitem__(self, item):
        from ..document import Document

        if isinstance(item, int):
            return Document(self._docs_proto[item])
        elif isinstance(item, str):
            return Document(self._docs_map[item])
        elif isinstance(item, slice):
            return DocumentSet(self._docs_proto[item])
        else:
            raise IndexError(f'do not support this index {item}')

    def __add__(self, other: 'DocumentSet'):
        v = DocumentSet([])
        for doc in self:
            v.add(doc)
        for doc in other:
            v.add(doc)
        return v

    def __iadd__(self, other: 'DocumentSet'):
        for doc in other:
            self.add(doc)
        return self

    def append(self, doc: 'Document') -> 'Document':
        """
        Append :param:`doc` in :class:`DocumentSet`.

        :param doc: The doc needs to be appended.
        :return: Appended list.
        """
        return self._docs_proto.append(doc.proto)

    def add(self, doc: 'Document') -> 'Document':
        """Shortcut to :meth:`append`, do not override this method.

        :param doc: the document to add to the set
        :return: Appended list.
        """
        return self.append(doc)

    def extend(self, iterable: Iterable['Document']) -> None:
        """
        Extend the :class:`DocumentSet` by appending all the items from the iterable.

        :param iterable: the iterable of Documents to extend this set with
        """
        for doc in iterable:
            self.append(doc)

    def clear(self):
        """Clear the data of :class:`DocumentSet`"""
        del self._docs_proto[:]

    def reverse(self):
        """In-place reverse the sequence."""
        if isinstance(self._docs_proto, RepeatedContainer):
            size = len(self._docs_proto)
            hi_idx = size - 1
            for i in range(int(size / 2)):
                tmp = DocumentProto()
                tmp.CopyFrom(self._docs_proto[hi_idx])
                self._docs_proto[hi_idx].CopyFrom(self._docs_proto[i])
                self._docs_proto[i].CopyFrom(tmp)
                hi_idx -= 1
        elif isinstance(self._docs_proto, list):
            self._docs_proto.reverse()

    def build(self):
        """Build a doc_id to doc mapping so one can later index a Document using doc_id as string key."""
        self._docs_map = {d.id: d for d in self._docs_proto}

    def sort(self, *args, **kwargs):
        """
        Sort the items of the :class:`DocumentSet` in place.

        :param args: variable set of arguments to pass to the sorting underlying function
        :param kwargs: keyword arguments to pass to the sorting underlying function
        """
        self._docs_proto.sort(*args, **kwargs)

    @property
    def all_embeddings(self) -> Tuple['np.ndarray', 'DocumentSet']:
        """Return all embeddings from every document in this set as a ndarray

        :return: The corresponding documents in a :class:`DocumentSet`,
                and the documents have no embedding in a :class:`DocumentSet`.
        :rtype: A tuple of embedding in :class:`np.ndarray`
        """
        return self._extract_docs('embedding')

    @property
    def all_contents(self) -> Tuple['np.ndarray', 'DocumentSet']:
        """Return all embeddings from every document in this set as a ndarray

        :return: The corresponding documents in a :class:`DocumentSet`,
                and the documents have no contents in a :class:`DocumentSet`.
        :rtype: A tuple of embedding in :class:`np.ndarray`
        """
        return self._extract_docs('content')

    def _extract_docs(
        self, *fields: str
    ) -> Tuple[Union['np.ndarray', List['np.ndarray']], 'DocumentSet']:

        list_of_contents_output = len(fields) > 1
        contents = [[] for _ in fields if len(fields) > 1]
        docs_pts = []
        bad_docs = []

        if list_of_contents_output:
            for doc in self:
                content = doc.get_attrs_values(*fields)
                if content is None:
                    bad_docs.append(doc)
                    continue
                for i, c in enumerate(content):
                    contents[i].append(c)
                docs_pts.append(doc)
            for i in range(len(contents)):
                contents[i] = np.stack(contents[i])
        else:
            for doc in self:
                content = doc.get_attrs_values(*fields)[0]
                if content is None:
                    bad_docs.append(doc)
                    continue
                contents.append(content)
                docs_pts.append(doc)
            contents = np.stack(contents) if contents else None

        if bad_docs:
            default_logger.warning(
                f'found {len(bad_docs)} docs at granularity {bad_docs[0].granularity} are missing one of the '
                f'following fields: {fields} '
            )

        return contents, DocumentSet(docs_pts)

    def __bool__(self):
        """To simulate ```l = []; if l: ...```

        :return: returns true if the length of the set is larger than 0
        """
        return len(self) > 0

    def new(self) -> 'Document':
        """Create a new empty document appended to the end of the set.

        :return: a new Document appended to the set
        """
        from ..document import Document

        return self.append(Document())

    def __str__(self):
        from ..document import Document

        content = ',\n'.join(str(Document(d)) for d in self._docs_proto[:3])
        if len(self._docs_proto) > 3:
            content += f'in total {len(self._docs_proto)} items'
        return content

    def __repr__(self):
        content = ' '.join(
            f'{k}={v}' for k, v in {'length': len(self._docs_proto)}.items()
        )
        content += f' at {id(self)}'
        content = content.strip()
        return f'<{typename(self)} {content}>'



import uuid
from typing import Optional

from fastapi import APIRouter, File, UploadFile, Body
from fastapi.exceptions import HTTPException

from ... import Runtime400Exception
from ...models import FlowModel
from ...models.status import FlowStoreStatus, FlowItemStatus
from ...stores import flow_store as store

router = APIRouter(prefix='/flows', tags=['flows'])


@router.get(
    path='', summary='Get all alive Flows\' status', response_model=FlowStoreStatus
)
async def _get_items():
    return store.status


@router.get(path='/arguments', summary='Get all accept arguments of a Flow')
async def _fetch_flow_params():
    return FlowModel.schema()['properties']


@router.post(
    path='',
    summary='Create a Flow from a YAML config',
    status_code=201,
    response_model=uuid.UUID,
)
async def _create(
    flow: UploadFile = File(...), workspace_id: Optional[uuid.UUID] = Body(None)
):
    try:
        return store.add(flow.file, workspace_id)
    except Exception as ex:
        raise Runtime400Exception from ex


# order matters! this must be put in front of del {id}
#  https://fastapi.tiangolo.com/tutorial/path-params/?h=+path#order-matters
@router.delete(
    path='',
    summary='Terminate all running Flows',
)
async def _clear_all():
    store.clear()


@router.delete(
    path='/{id}',
    summary='Terminate a running Flow',
    description='Terminate a running Flow and release its resources',
)
async def _delete(id: uuid.UUID, workspace: bool = False):
    try:
        store.delete(id=id, workspace=workspace)
    except KeyError:
        raise HTTPException(status_code=404, detail=f'{id} not found in {store!r}')


@router.get(
    path='/{id}',
    summary='Get the status of a running Flow',
    response_model=FlowItemStatus,
)
async def _status(
    id: 'uuid.UUID',
):
    try:
        return store[id]
    except KeyError:
        raise HTTPException(status_code=404, detail=f'{id} not found in {store!r}')


@router.on_event('shutdown')
def _shutdown():
    store.reset()



import numpy as np
from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders import BaseAudioEncoder


class MFCCTimbreEncoder(BaseAudioEncoder):
    """
    Extract a `n_mfcc`-dimensional feature vector for each MFCC frame.

    :class:`MFCCTimbreEncoder` is based on Mel-Frequency Cepstral
        Coefficients (MFCCs) which represent timbral features.
    :class:`MFCCTimbreEncoder` encodes an audio signal from a
        `Batch x Signal Length` ndarray into a
        `Batch x Concatenated Features` ndarray.

    :param input_sample_rate: input sampling rate in Hz
        (22050 by default)
    :param n_mfcc: the number of coefficients
        (20 by default)
    :param n_fft: length of the FFT window
        (2048 by default)
    :param hop_length: the number of samples between
        successive MFCC frames (512 by default)
    """

    def __init__(self, input_sample_rate: int = 22050, n_mfcc: int = 20, n_fft_length: int = 2048,
                 hop_length: int = 512, *args, **kwargs):
        """Set Constructor."""
        super().__init__(*args, **kwargs)
        self.input_sample_rate = input_sample_rate
        self.n_mfcc = n_mfcc
        self.n_fft_length = n_fft_length
        self.hop_length = hop_length

    @batching
    @as_ndarray
    def encode(self, data: np.ndarray, *args, **kwargs) -> np.ndarray:
        """
        Craft the audio signal of each Document into short MFCC frames.

        Extract MFCCs for each frame and concatenates Document frame
            MFCCs into a single Document embedding.

        :param data: a `Batch x Signal Length` ndarray,
            where `Signal Length` is a number of samples
        :return: a `Batch x Concatenated Features` ndarray,
            where `Concatinated Features` is a `n_mfcc`-dimensional
            feature vector times the number of the MFCC frames
        """
        from librosa.feature import mfcc
        embeds = []
        for chunk_data in data:
            mfccs = mfcc(y=chunk_data, sr=self.input_sample_rate, n_mfcc=self.n_mfcc, n_fft=self.n_fft_length,
                         hop_length=self.hop_length)
            embeds.append(mfccs.flatten())
        return embeds



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np
import torch
import clip

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTorchEncoder
from jina.executors.devices import TorchDevice

class CLIPTextEncoder(BaseTorchEncoder):
    """Encode data from a `np.ndarray` (of strings) of length `BatchSize` into
    a `np.ndarray` of shape `Batchsize x EmbeddingDimension`. 

    Internally, :class:`CLIPImageEncoder` wraps the `CLIP` model from `https://github.com/openai/CLIP`.

    :param model_name: The name of the model. Supported models include ``ViT-B/32`` and ``RN50``.
    :param args: Additional positional arguments.
    :param kwargs: Additional positional arguments.
    """
    def __init__(self, model_name: str ='ViT-B/32',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_name = model_name

    def post_init(self):
        """Load a model from clip specified in `model_name`. """
        import clip
        model, _ = clip.load(self.model_name, self.device)
        self.model = model

    @batching
    @as_ndarray
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """Transform a `np.ndarray` of strings of length `BatchSize` into
        a `np.ndarray` of shape `Batchsize x EmbeddingDimension`.

        :param data: A `np.ndarray` of strings.
        :param args: Additional positional arguments.
        :param kwargs: Additional positional arguments.
        :return: A `BachSize x EmbeddingSize` numpy `ndarray`.
        """
        input_torchtensor = clip.tokenize(data)
        if self.on_gpu:
            input_torchtensor = input_torchtensor.cuda()

        with torch.no_grad():
            embedded_data = self.model.encode_text(input_torchtensor)

        embedded_data = embedded_data.cpu().numpy()
        return embedded_data
    



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple, Dict, Union

import numpy as np
from jina.executors.crafters import BaseCrafter

from .helper import _crop_image, _move_channel_axis, _load_image


class CenterImageCropper(BaseCrafter):
    """
    Crop the image with the center crop box.

    The coordinate is the same coordinate-system in the
        :py:mode:`PIL.Image`.

    :param target_size: Desired output size. If size
        is a sequence like (h, w), the output size will
        be matched to this. If size is an int, the
        output will have the same height and width as
        the `target_size`.
    :param channel_axis: Axis for channel
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self,
                 target_size: Union[Tuple[int, int], int] = 224,
                 channel_axis: int = -1,
                 *args,
                 **kwargs):
        """Set Constructor."""
        super().__init__(*args, **kwargs)
        self.target_size = target_size
        self.channel_axis = channel_axis

    def craft(self, blob: 'np.ndarray', *args, **kwargs) -> Dict:
        """
        Crop the input image array.

        :param blob: The ndarray of the image
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        :return: A dict with the cropped image
        """
        raw_img = _load_image(blob, self.channel_axis)
        _img, top, left = _crop_image(raw_img, self.target_size, how='center')
        img = _move_channel_axis(np.asarray(_img), -1, self.channel_axis)
        return dict(offset=0, blob=img.astype('float32'), location=(top, left))



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict

import numpy as np
from jina.executors.rankers import Chunk2DocRanker


class BM25Ranker(Chunk2DocRanker):
    """
    :class:`BM25Ranker` calculates the weighted score from the matched chunks.
        The weights of each chunk is based on the tf-idf algorithm.
        Each query chunk is considered as a ``term``,
        and the frequency of the query chunk in a specific matched document
        is considered as the naive ``term-frequency``.
        All the matched results as a whole are considered as the corpus,
        and therefore the frequency of the query chunk in all the matched docs
        is considered as the naive ``document-frequency``.
        Please refer to the functions for the details
        of calculating ``tf`` and ``idf``.

    :param k: the parameter ``k`` for the BM25 algorithm.
    :param b: the parameter ``b`` for the BM25 algorithm.
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """
    required_keys = {'length', }

    def __init__(self, k=1.2, b=0.75, threshold=0.1, *args, **kwargs):
        """Set constructor."""
        super().__init__(*args, **kwargs)
        self.threshold = threshold
        self.k = k
        self.b = b

    def score(self, match_idx: 'np.ndarray', query_chunk_meta: Dict, match_chunk_meta: Dict) -> 'np.ndarray':
        """
        Calculate the score from matched chunks

        :param match_idx: an `ndarray` of the size ``N x 4``. `
            `N`` is the batch size of the matched chunks for the query doc.
            The columns correspond to the ``doc_id`` of the matched chunk,
            ``chunk_id`` of the matched chunk,
            ``chunk_id`` of the query chunk, and ``score`` of the matched chunk.
        :param query_chunk_meta: a dict of meta info for the query chunks
            with **ONLY** the ``required_keys`` are kept.
        :param match_chunk_meta: a dict of meta info for the matched chunks
            with **ONLY** the ``required_keys`` are kept.

        :return: an `ndarray` of the size ``M x 2``.
            ``M`` is the number of matched docs.
            The columns correspond to the ``doc_id`` and ``score``.

        .. note::
            In both `query_chunk_meta` and `match_chunk_meta`,
            ONLY the fields from the ``required_keys`` are kept.
        """
        _groups = self.group_by_doc_id(match_idx)
        r = []
        _q_idf = self.get_idf(match_idx)
        for _g in _groups:
            _doc_id, _doc_score = self._get_score(_g, query_chunk_meta, match_chunk_meta, _q_idf)
            r.append((_doc_id, _doc_score))
        return self.sort_doc_by_score(r)

    def _get_df(self, match_idx):
        """Get the naive document frequency.

        :param match_idx: an `ndarray` of the size ``N x 4``.
            ``N`` is the number of chunks in a given doc that matched
            with the query doc.

        :return: a tuple of two `np.ndarray` in the size of ``M``,
            i.e. the document frequency array and the chunk id
            array. ``M`` is the number of query chunks.
        """
        a = match_idx[match_idx[self.COL_DOC_CHUNK_ID].argsort()]
        q_id, q_df = np.unique(a[self.COL_DOC_CHUNK_ID], return_counts=True)
        return q_df, q_id

    def _get_tf(self, match_idx):
        """Get the naive term frequency of the query chunks.

        :param match_idx: an `ndarray` of the size ``N x 4``.
            ``N`` is the number of chunks in a given doc that matched
            with the query doc.
        :return: a tuple of three `np.ndarray` in the size of ``M``,
            i.e. the term frequency array, the query chunk id array,
            and the matched chunk id array.``M`` is the number of query chunks.

        .. note::
            The query chunks with matching scores
            that is lower than the threshold are dropped.
        """
        _m = match_idx[match_idx[self.COL_SCORE] >= self.threshold]
        _sorted_m = _m[_m[self.COL_DOC_CHUNK_ID].argsort()]
        q_id_list, q_tf_list = np.unique(_sorted_m[self.COL_DOC_CHUNK_ID], return_counts=True)
        row_id = np.cumsum(q_tf_list) - 1
        c_id_list = _sorted_m[row_id][self.COL_MATCH_ID]
        return q_tf_list, q_id_list, c_id_list

    def get_idf(self, match_idx):
        """Get the idf dictionary for query chunks that matched a given doc.

        :param match_idx: an `ndarray` of the size ``N x 4``.
            ``N`` is the batch size of the matched chunks for the query doc.
            The columns correspond to the ``doc_id`` of the matched chunk,
            ``chunk_id`` of the matched chunk,
             ``chunk_id`` of the query chunk, and ``score`` of the matched chunk.
        :return: a dict in the size of query chunks

        .. note::
            The 10-based logarithm version idf is used,
                i.e. idf = log10(1 + (total - df + 0.5) / (df + 0.5)).
                ``df`` denotes the frequency of the query chunk in all the matched chunks.
                `total` denotes the total number of the matched chunks.
        """
        _q_df, _q_id = self._get_df(match_idx)
        _total_df = np.sum(_q_df)
        return {idx: np.log10((_total_df + 1.) / (df + 0.5)) ** 2 for idx, df in zip(_q_id, _q_df)}

    def get_tf(self, match_idx, match_chunk_meta):
        """Get the tf dictionary for query chunks that matched a given doc.

        :param match_idx: an `ndarray` of the size ``N x 4``.
            ``N`` is the number of chunks in a given doc
            that matched  with the query doc.
        :param match_chunk_meta: a dict of meta info for
            the matched chunks with **ONLY**
            the ``required_keys`` are kept.

        :return: a dict in the size of query chunks
        .. note::
            The term-frequency of a query chunk is frequency
                of the query chunk that has a matching score equal or
                higher than the ``threshold``. In BM25:
                tf = (1 + k) * tf / (k * (1 - b + b * n_doc / avg_n_doc) + tf).
                ``n`` denotes thefrequency of the query chunk in the matched doc.
                ``n_doc`` denotes the total number of chunks in the matched doc.
                ``avg_n_doc`` denotes the average number
                of chunks over all the matched docs.
        """
        _q_tf_list, _q_id_list, _c_id_list = self._get_tf(match_idx)
        _avg_n_doc = np.mean([c_meta['length'] for c_meta in match_chunk_meta.values()])
        return {q_idx: (1 + self.k) * tf / (
                self.k * (1 - self.b + self.b * match_chunk_meta[c_idx]['length'] / _avg_n_doc) + tf)
                for c_idx, q_idx, tf in zip(_c_id_list, _q_id_list, _q_tf_list)}

    def _get_score(self, match_idx, query_chunk_meta, match_chunk_meta, idf, *args, **kwargs):
        """Get the doc score based on the weighted sum of matching scores.
        The weights are calculated from the tf-idf of the query chunks.

        :param match_idx: an `ndarray` of the size ``N x 4``.
            ``N`` is the number of chunks in a given doc that matched
            with the query doc.
        :param tf: a dictionary with the query chunk id as key and the tf as value.
        :param idf: a dictionary with the query chunk id as key and the idf as value.

        :return: a scalar value of the weighted score.
        """
        tf = self.get_tf(match_idx, match_chunk_meta)
        _weights = match_idx[self.COL_SCORE]
        _q_tfidf = np.vectorize(tf.get)(match_idx[self.COL_DOC_CHUNK_ID], 0) * \
                   np.vectorize(idf.get)(match_idx[self.COL_DOC_CHUNK_ID], 0)
        _sum = np.sum(_q_tfidf)
        _doc_id = self.get_doc_id(match_idx)
        _score = 0. if _sum == 0 else np.sum(_weights * _q_tfidf) * 1.0 / _sum
        return _doc_id, _score
